# -*- coding: utf-8 -*-

import logging

from artixds.objectdiscount import PositionObject
from artixds.impact import Impact, ImpactType, ImpactObjectPosition, ImpactMode
from artixds.branchdetail import BranchDetail
from artixds.discountrate import DiscountRateType
from artixds.config import Config

from kitobject import KitObject, KitObjectItem
from klkititem import KlKitItem
from klposition import KlPosition
from klpositionpart import KlPositionPart

log = logging.getLogger("KitLogic")

class KitLogic(object):
    '''
    Новыя логика для работы с наборами

    При помощи метода findKit ищется набор в чеке и если набор найден возвращается true.
    После чего из полей structureImpacts, pureImpacts, uniteImpacts можно получить воздействия,
    а в поле partsOfKit список номеров позиций которые входят в набор.

    Если прям совсем непонятно как работает, вот тут немного понятнее описано
    http://docs.artix.su/x/EwbY
    '''
    def __init__(self, check, kit, discRate, campaign, discount):
        self.check = check
        self.kit = kit
        self.discRate = discRate
        self.nextPosNum = check.getNextPositionNumber()
        self.campaign = campaign
        self.discount = discount

        self.pureImpacts = []   # Скидочные воздействия в которых нет номера позиции
        self.pureImpactsWithoutNum = []
        self.uniteImpacts = []
        self.structureImpacts = []

    def getNextPosNum(self):
        '''
        Возвращает номер позиции для бранча
        '''
        result = self.nextPosNum
        self.nextPosNum += 1
        return result

    def distributeForKitItems(self, positions):
        '''
        Раскидать позиции по частям набора

        Если позиция подходит в оба элемента набора, позиция добавляется в каждый
        в виде KlPositionPart

        Позиции которые используются в других наборах и расфасованные игнорируются
        '''
        for pos in positions:
            for i in range(len(self.posPerKitItem)):
                kitItem = self.posPerKitItem[i].kitItem
                if self.posPerKitItem[i].fract and not pos.fract:
                    log.warn(u"Не весовая позиция %s не может участвовать в части набора с дробным кол-вом %s" % (pos, kitItem.name))
                    continue

                posCatalogs = set(map(lambda x: x["catalogid"], pos.pos.data['tmc']['catalogs']))
                if set(kitItem.catalogCodes) & posCatalogs:
                    if not pos.usedTimes():
                        pos.kitItemNum = i
                        self.posPerKitItem[i].append(pos)
                    elif pos.usedTimes() > 1:
                        self.posPerKitItem[i].append(pos.createPart())
                    else:
                        self.posPerKitItem[pos.kitItemNum].remove(pos)
                        self.posPerKitItem[pos.kitItemNum].append(pos.createPart())
                        self.posPerKitItem[i].append(pos.createPart())

    def applyDisc(self, pos, kitItem):
        '''
        Применение скидочного воздействия на позицию pos
        '''
        isPart = isinstance(pos, KlPositionPart)
        klPos = pos.source()

        if isPart:
            # сохраняем разбитые позиции, чтобы потом сделать структурные импакты
            self.posWithBranches.add(klPos)
        else:
            # сохраняем номера позиций которые стали частью набора
            # у части позиции еще нет номера, потом сохраним
            self.partsOfKit.add(pos.posnum)
        pos.partOfKit = True

        if kitItem.discRate:
            rateValue = kitItem.discRate.getAbsoluteValue(PositionObject(klPos.pos.data, pos), self.check)
            # Если суммовая скидка,
            if kitItem.discRate.getRateType() == DiscountRateType.SUM:
                rateValue = rateValue * pos.bquant / kitItem.kitItem.quantity

            impact = Impact(type=ImpactType.PURE,
                campaign=self.campaign,
                discount=self.discount,
                absoluteValue=rateValue,
                impactObject=ImpactObjectPosition(pos.posnum, code=klPos.pos.data['code']),
                qty=pos.bquant,
                mode = ImpactMode.BYDISCKIT,
                discountRate=kitItem.discRate)

            self.pureImpacts.append(impact)
            if isPart:
                # у части позиции еще нет номера, потом установим
                self.pureImpactsWithoutNum.append((impact, pos))

    def distributeCommon(self):
        '''
        Раскидываение общих позиций между частями набора
        '''
        if any([ki.hasCommon for ki in self.posPerKitItem]):
            # Считаем общее количество в позициях которые входят в пересекающихся частях набора
            commonPosSet = set()
            commonPosCount = sum([ki.uniqueCount(commonPosSet) for ki in self.posPerKitItem if ki.hasCommon])
            # Считаем общее количество весов (вес это количество элементов в части набора)
            weighSum = sum([ki.kitItem.quantity for ki in self.posPerKitItem if ki.hasCommon])
            # коэффициент распределения общих позиций (его умножаем на кол-во в элементе набора
            # и получаем кол-во позиций для оптимального кол-ва вхождения набора)
            factor = commonPosCount / weighSum

            # Распределение общих позиций
            for kitItem in self.posPerKitItem:
                kitItem.distributeCommon(factor)

    def calcOccurrences(self):
        '''
        Считает сколько раз набор входит в чек
        '''
        occurrenceCount = min([int(round(ki.count(), 3) / ki.kitItem.quantity) for ki in self.posPerKitItem])
        if self.kit.maxtimes:
            occurrenceCount = min(occurrenceCount, self.kit.maxtimes)
        return occurrenceCount

    def selectPositions(self, occurrenceCount):
        '''
        Выделяет в каждой части набора позиции которые попали в набора для
        заданного количества вхождений

        Может объединить позиции. Если часть позиции занимает все количество исходной,
        из нее сделает полную. От полной может отрезать часть.
        '''

        unitePositionsInKit = Config().getBool('Discounts', 'unitePositionsInKit', True)
        log.debug(u"Объединение позиций " + u"включено" if unitePositionsInKit else u"выключено")

        for kitItem in self.posPerKitItem:
            neededCount = round(occurrenceCount * kitItem.kitItem.quantity, 3)
            # позиция в которую объединяются другие
            collector = None
            for i in range(len(kitItem.positions)):
                pos = kitItem.positions[i]
                isPart = isinstance(pos, KlPositionPart)

                if neededCount < 0.001:
                    if isPart:
                        pos.setQuantity(0)
                    continue

                klPos = pos.source()
                if isPart:
                    # Если есть что-то неиспользуемое в исходной позиции засунем сюда
                    pos.setQuantity(pos.bquant + klPos.restCount())

                if pos.bquant < 0.001:
                    continue

                # определяемся с количеством которое заберем у позиции
                if neededCount >= pos.bquant:
                    subCount = pos.bquant
                    if isPart and pos.isWholePos():
                        kitItem.positions[i] = klPos
                        pos = klPos
                        isPart = False
                else:
                    subCount = neededCount
                    if not isPart:
                        pos = klPos.createPart()
                        isPart = True
                    pos.setQuantity(subCount)
                neededCount = round(neededCount - subCount, 3)

                # объединение позиций
                if unitePositionsInKit:
                    posKey = klPos.uniqueKey()
                    if not collector:
                        collector = pos
                    elif collector.uniqueKey() != posKey:
                        self.applyDisc(collector, kitItem)
                        collector = pos
                    else:
                        collector.uniteWith(pos)
                        self.united.add(collector.source())
                else:
                    self.applyDisc(pos, kitItem)

            if unitePositionsInKit:
                self.applyDisc(collector, kitItem)

    def findKit(self):
        log.debug(u"Поиск вхождений набора '%s' в чек" % self.kit.name)

        # Обворачиваем позиции
        positions = [KlPosition(pos) for pos in self.check.getGoodsList() if not pos.data["partOfKit"] and not pos.data["packingMode"]]

        if not positions:
            return False

        if hasattr(self.kit, 'byTmc') and self.kit.byTmc:
            # получаем множество кодов товаров
            codeSet = set([p.pos.data['code'] for p in positions])

            # для каждого из кодов товара ищем наборы
            result = False
            for code in codeSet:
                result |= self.makeKit(filter(lambda p: p.pos.data['code'] == code, positions))
            return result

        return self.makeKit(positions)


    def makeKit(self, positions):
        self.united = set()     # объединенные позиции
        self.posWithBranches = set()
        self.partsOfKit = set()

        # Элементы набора
        posPerKitItem = [KlKitItem(ki, self.discRate.discountRateForItem(ki), self.check) for ki in self.kit]

        # Пересортируем элементы набора так, чтобы первыми были части со скидками
        self.posPerKitItem = sorted(posPerKitItem, key = lambda x: x.discRate == None)

        # Раскидываем позиции по частям набора
        self.distributeForKitItems(positions)

        # Сортируем позиции в частях набора
        for kitItem in self.posPerKitItem:
            kitItem.sort()

        # Делим между частями общие позиции
        self.distributeCommon()

        log.debug(u"Распределение позиций по частям набора - %s" % self.posPerKitItem)

        # Считаем количество вхождений набора
        occurrenceCount = self.calcOccurrences()
        log.debug(u"Количество вхождений набора - %s" % occurrenceCount)
        if occurrenceCount < 1:
            return False

        # Выделим позиции которые входят в набора
        self.selectPositions(occurrenceCount)

        # Создаем импакты объединения
        for pos in self.united:
            self.uniteImpacts.append(Impact.uniteImpact(
                ImpactObjectPosition(pos.posnum, pos.pos.data['code']), pos.bquant, pos.unitedPositions))

        # Создаем структурные импакты
        # Сортируем чтобы новые позиции были с номерами по порядку
        for pos in sorted(self.posWithBranches, key=lambda x: x.posnum):
            pos.complete(self.getNextPosNum)
            if not pos.parts:
                # Если нет частей, значит объединили
                self.partsOfKit.add(pos.posnum)
                continue
            self.structureImpacts.append(Impact.branchImpact(
                    ImpactObjectPosition(pos.posnum, pos.pos.data['code']),
                    [BranchDetail(i.posnum, i.bquant, i.sumn, i.sumb, i.partOfKit) for i in pos.parts]))
            self.partsOfKit.update([i.posnum for i in pos.parts if i.partOfKit])

        # Не забыли про импакты с частями позиций у которых небыло номеров
        for impact, pos in self.pureImpactsWithoutNum:
            if pos.posnum == None:
                # Если нет номера, значит объединили
                impact.impactObject.positionNumber = pos.source().posnum
            else:
                impact.impactObject.positionNumber = pos.posnum

        log.debug(u"Воздействия объединения - %s" % self.uniteImpacts)
        log.debug(u"Cтруктурные воздействия - %s" % self.structureImpacts)
        log.debug(u"Скидочные воздействия - %s" % self.pureImpacts)
        log.debug(u"Номера позиций которые попали в набор - %s" % self.partsOfKit)

        return True

